/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Wez Furlong <wez@php.net>                                    |
  +----------------------------------------------------------------------+
*/

/* $Id: win32.c,v 1.2 2004/12/16 05:31:33 wez Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_event.h"
#include "php_event_int.h"
#include "php_network.h"

/* win32 engine.
 * On windows we have an obligation to process and dispatch
 * messages on a GUI thread.  That means that we should never
 * truly block when waiting for something to happen, as it
 * could cause a deadlock, particulary for scripts that use
 * libraries that create window handles or scripts that use
 * COM (which requires a working message pump).
 */

struct win32_item {
	struct php_event *ev;
};

struct win32_queue {
	HashTable events;
	HWND select_window;
	void ***tsrm_ls;
	HANDLE fake_event;
	int got_events;
};

static void item_dtor(void *pDest)
{
	struct win32_item *item = (struct win32_item*)pDest;
	if (item->ev) {
		WSAAsyncSelect(item->ev->ev_fd, NULL, 0, 0);
	}
}

#ifdef COMPILE_DL_EVENT
static HINSTANCE extension_instance = NULL;

BOOL WINAPI DllMain(HINSTANCE inst, DWORD reason, LPVOID reserved)
{
	if (reason == DLL_PROCESS_ATTACH) {
		extension_instance = inst;
	}
	return TRUE;
}
#else
# define extension_instance GetModuleHandle(NULL)
#endif

static LRESULT CALLBACK select_wnd_proc(HWND hwnd, UINT uMsg,
	WPARAM wParam, LPARAM lParam)
{
	struct win32_queue *q;
	struct win32_item *item;
	void ***tsrm_ls;

	switch (uMsg) {
		case WM_USER+1:
			/* we got an event notification */
			q = (struct win32_queue*)GetWindowLongPtr(hwnd, GWL_USERDATA);

			if (!q) return 1;
			tsrm_ls = q->tsrm_ls; /* fake access to correct TLS */

			if (SUCCESS == zend_hash_index_find(&q->events, (ulong)wParam, &item)) {
				short mask = 0;
				WORD wmask = LOWORD(lParam);

				if (wmask & FD_READ)
					mask |= EV_READ;
				if (wmask & FD_WRITE)
					mask |= EV_WRITE;
				if (wmask & FD_OOB)
					mask |= EV_EXCEP;

				if (item->ev) {
					event_active(item->ev, mask, 1 TSRMLS_CC);
					q->got_events = 1;
				}
			}
			return 0;
	}

	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

static void *win32_event_init(TSRMLS_D) {
	WNDCLASS cls;
	struct win32_queue *q;

	q = emalloc(sizeof(*q));
	zend_hash_init(&q->events, 32, NULL, item_dtor, 0);

	memset(&cls, 0, sizeof(cls));
	cls.lpfnWndProc = select_wnd_proc;
	cls.lpszClassName = "PHP_PECL_event_select_window";
	cls.hInstance = extension_instance;
	RegisterClass(&cls);

	q->select_window = CreateWindow(
		cls.lpszClassName,
		cls.lpszClassName,
		0,
		0, 0, 0, 0, 
		HWND_MESSAGE,
		NULL,
		extension_instance,
		NULL);
	
	q->tsrm_ls = tsrm_ls;
	SetWindowLongPtr(q->select_window, GWL_USERDATA, (LONG_PTR)q);

	q->fake_event = CreateEvent(NULL, TRUE, FALSE, NULL);

	
	return q;
}

static void win32_event_fini(void *arg TSRMLS_DC)
{
	struct win32_queue *q = arg;

	DestroyWindow(q->select_window);
	zend_hash_destroy(&q->events);
	efree(q);
}

static int win32_event_recalc(void *arg, int max TSRMLS_DC)
{
	struct win32_queue *q = arg;
	return 0;
}

static int win32_event_add(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct win32_queue *q = arg;
	
	if (ev->php_stream && Z_TYPE_P(ev->php_stream) == IS_RESOURCE) {
		php_stream *stream;
		php_stream_from_zval_no_verify(stream, &ev->php_stream);
		if (php_stream_is(stream, PHP_STREAM_IS_SOCKET)) {
			struct win32_item item;
			DWORD mask = 0;
			
			memset(&item, 0, sizeof(item));
			item.ev = ev;

			if (ev->ev_events & EV_READ) {
				mask |= FD_READ|FD_CLOSE;
			}
			if (ev->ev_events & EV_WRITE) {
				mask |= FD_WRITE;
			}
			if (ev->ev_events & EV_EXCEP) {
				mask |= FD_OOB;
			}
			WSAAsyncSelect(ev->ev_fd, q->select_window,
				WM_USER+1, mask);

			zend_hash_index_update(&q->events, ev->ev_fd, &item, sizeof(item), NULL);
			
			return 0;
		}
	}

	/* otherwise we could initiate an overlapped read/write so that
	 * we get notified when its done, but it probably won't work 
	 * unless the file was opened overlapped */

	return -1;
}

static int win32_event_del(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct win32_queue *q = arg;
	zend_hash_index_del(&q->events, ev->ev_fd);
	return 0;
}

static int win32_event_dispatch(void *arg, struct timeval *tv TSRMLS_DC)
{
	struct win32_queue *q = arg;
	DWORD ms_total, limit, ret;
	struct php_event *ev;
	int n_handles;

	if (tv == NULL) {
		ms_total = INFINITE;
	} else {
		ms_total = tv->tv_sec * 1000;
		ms_total += tv->tv_usec / 1000;
		limit = GetTickCount() + ms_total;
	}

	q->got_events = 0;
	do {
		if (ms_total != INFINITE) {
			ms_total = limit - GetTickCount();
		}
		ret = MsgWaitForMultipleObjectsEx(1, &q->fake_event,
			ms_total, QS_ALLEVENTS|QS_ALLINPUT|QS_ALLPOSTMESSAGE|
			QS_MOUSE|QS_SENDMESSAGE|QS_TIMER,
			MWMO_INPUTAVAILABLE|MWMO_ALERTABLE);

		if (ret == WAIT_OBJECT_0+1) {
			/* got some message(s) to process */
			MSG m;

			while (PeekMessage(&m, NULL, 0, 0, PM_REMOVE)) {
				TranslateMessage(&m);
				DispatchMessage(&m);
			}
		} else if (ret == WAIT_TIMEOUT) {
			break;
		}
	} while (!q->got_events && (ms_total == INFINITE || GetTickCount() < limit));

	return 0;
}

const struct php_eventop php_event_win32_engine = {
	"win32",
	win32_event_init,
	win32_event_fini,
	win32_event_add,
	win32_event_del,
	win32_event_recalc,
	win32_event_dispatch
};


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
